from typing import Any


class Serializer:
    def serialize(self, x : Any) -> bytes:
        raise NotImplementedError
    
    def deserialize(self, x : bytes) -> Any:
        raise NotImplementedError


import argparse
from .bind import bind_obj, bind_row
from ..version import version

def print_version(args):
    print("""KARA Storage version `{version}`.
Website: https://git.thunlp.vip/kara/kara-row-storage""".format(version=version))

def get_parser():
    parser = argparse.ArgumentParser(prog="kara_storage", description="KARA Storage command line interface.")
    parser.set_defaults(func=None)
    sub_parsers = parser.add_subparsers(help="storage types")
    
    bind_obj(sub_parsers.add_parser("obj", help="object storage"))
    bind_obj(sub_parsers.add_parser("object", help="object storage"))
    bind_row(sub_parsers.add_parser("row", help="row storage"))
    version = sub_parsers.add_parser("version", help="print kara_storage version")
    version.set_defaults(func=print_version)
    return parser


def main():
    parser = get_parser()

    args = parser.parse_args()
    if args.func is None:
        parser.print_help()
    else:
        args.func(args)


import multiprocessing
multiprocessing.set_start_method("spawn", force=True)
# FIXME: force using spawn start method, use register_after_fork in the future.

from .storage import KaraStorage
from . import serialization
from .pytorch import make_torch_dataset
from .version import version
from .row import RowDataset


from typing import Any
from ..abc import Serializer
import pickle

class PickleSerializer(Serializer):
    def serialize(self, x : Any) -> bytes:
        return pickle.dumps(x)
    
    def deserialize(self, x : bytes) -> Any:
        return pickle.loads(x)


import io
from typing import Optional
from ..abc import StorageBase
class TrunkController(io.RawIOBase):
    def __init__(self, storage : StorageBase, prefix : str, mode : str, max_file_size : int = 128 * 1024 * 1024) -> None:
        if mode != "r" and mode != "w":
            raise ValueError("Unknown mode `%s`" % mode)
        
        # normalize prefix
        if not prefix.endswith("/"):
            prefix = prefix + "/"
        
        # store parameters
        self.__storage = storage
        self.__prefix = prefix
        self.__max_file_size = max_file_size
        self.__closed = False

        # initialize mode
        self.__readable = ("r" in mode)
        self.__writable = ("w" in mode)

        # initialize file pointer
        self.__tell = 0
        self.__size = 0
        self.__file_sizes = []
    
        # calc trunks

        self.__num_trunks = 0
        while True:
            file_size = self.__storage.filesize( self.__prefix + "%d.blk" % self.__num_trunks )
            if file_size is None:
                # returns None if file not exists
                break
            
            # else
            self.__file_sizes.append(file_size)
            self.__num_trunks += 1
        
        self.__size += sum(self.__file_sizes)

        if self.__readable:
            if self.__num_trunks == 0:
                raise RuntimeError("Empty dataset !")
            self.__fp_read = self.__storage.open(self.__prefix + "0.blk", "r")
            self.__infile_offset = 0
            self.__curr_file = 0
        if self.__writable:
            if self.__num_trunks == 0:
                self.__file_sizes.append(0)
                self.__num_trunks += 1

            self.__fp_write = self.__storage.open(self.__prefix + "%d.blk" % (self.__num_trunks - 1), "a")
            self.__infile_offset = self.__file_sizes[-1]
            self.__curr_file = self.__num_trunks - 1
    
    def readable(self) -> bool:
        return self.__readable
    
    def writable(self) -> bool:
        return self.__writable
    
    def seekable(self) -> bool:
        return self.__readable

    def readinto(self, __buffer) -> Optional[int]:
        if not self.__readable:
            raise RuntimeError("Dataset not readable")
        if self.__closed:
            raise RuntimeError("Dataset is closed")

        lw = self.__fp_read.readinto(__buffer)

        # If the object is in non-blocking mode and no bytes are available, None is returned.
        # https://docs.python.org/3/library/io.html#io.RawIOBase
        if lw is None:
            return None

        self.__tell += lw
        self.__infile_offset += lw

        if self.__infile_offset == self.__file_sizes[self.__curr_file]:
            # reached the end of current file
            if self.__curr_file + 1 < self.__num_trunks:
                # open next trunk
                self.__fp_read.close()
                self.__curr_file += 1
                self.__infile_offset = 0
                self.__fp_read = self.__storage.open(self.__prefix + "%d.blk" % self.__curr_file, "r")
        else:
            if lw == 0:
                # __infile_offset < __file_size[__curr_file] but got EOF
                raise RuntimeError("Faile size not aligned: expected %d more bytes" % (self.__file_sizes[self.__curr_file] - self.__infile_offset))
        return lw
    
    def write(self, __b : bytes) -> Optional[int]:
        if not self.__writable:
            raise RuntimeError("Dataset not writable")
        if self.__closed:
            raise RuntimeError("Dataset is closed")

        wrt_len = min( len(__b), self.__max_file_size - self.__infile_offset )
        self.__fp_write.append( __b[:wrt_len] )

        self.__size += wrt_len
        self.__infile_offset += wrt_len

        if self.__infile_offset == self.__max_file_size:
            self.__fp_write.close()
            self.__fp_write = self.__storage.open( self.__prefix + "%d.blk" % self.__num_trunks, "a")
            self.__num_trunks += 1
            self.__infile_offset = 0
        return wrt_len
    
    def seek(self, __offset: int, __whence: int = io.SEEK_SET) -> int:
        if not self.__readable:
            raise RuntimeError("Dataset not readable")
        if self.__closed:
            raise RuntimeError("Dataset is closed")

        nw_pos = __offset
        if __whence == io.SEEK_CUR:
            nw_pos = self.__tell + __offset
        elif __whence == io.SEEK_END:
            nw_pos = self.__size - __offset
        if nw_pos < 0:
            nw_pos = 0
        if nw_pos > self.__size:
            nw_pos = self.__size
        
        rest_size = nw_pos
        self.__curr_file = 0

        while rest_size > self.__file_sizes[self.__curr_file]:
            rest_size -= self.__file_sizes[self.__curr_file]
            self.__curr_file += 1
        
        # 处理最后相等的情况
        if rest_size == self.__file_sizes[self.__curr_file]:
            # 处理最后一个文件的情况
            if self.__curr_file + 1 < self.__num_trunks:
                self.__curr_file += 1
                rest_size = 0

        self.__fp_read.close()
        if rest_size == self.__file_sizes[self.__curr_file]:
            self.__fp_read = io.BytesIO()
        else:
            self.__fp_read = self.__storage.open(self.__prefix + "%d.blk" % self.__curr_file, "r", begin=rest_size)
        self.__tell = nw_pos

        return self.__tell

    def flush(self) -> None:
        if not self.__writable:
            return # ignore flush
        if self.__closed:
            raise RuntimeError("Dataset is closed")
        self.__fp_write.flush()
    
    def tell(self) -> int:
        return self.__tell
    
    def close(self):
        if not self.__closed:
            if self.__readable:
                self.__fp_read.close()
                self.__fp_read = None
            if self.__writable:
                self.__fp_write.close()
                self.__fp_write = None
            self.__closed = True
    
    @property
    def closed(self) -> bool:
        return self.__closed
    
    @property
    def size(self) -> int:
        return self.__size

    def pread(self, offset : int, length : int) -> bytes:
        if not self.__readable:
            raise RuntimeError("Dataset not readable")
        if self.__closed:
            raise RuntimeError("Dataset is closed")

        trunk_id = 0
        while offset >= self.__file_sizes[trunk_id] and trunk_id < self.__num_trunks:
            offset -= self.__file_sizes[trunk_id]
            trunk_id += 1
        
        if trunk_id >= self.__num_trunks:
            return b""
        
        rest_length = length
        
        ret = bytearray(length)
        read_offset = 0
        view = memoryview(ret)


        while rest_length > 0 and trunk_id < self.__num_trunks:
            ed = self.__file_sizes[trunk_id]
            if offset + rest_length < ed:
                ed = offset + rest_length
            fp = self.__storage.open(self.__prefix + "%d.blk" % trunk_id, "r", offset, ed)
            offset = ed % self.__file_sizes[trunk_id]

            
            vlen = fp.readinto(view[read_offset:])

            read_offset += vlen
            rest_length -= vlen
            trunk_id += 1
            fp.close()
        return bytes( ret[:read_offset] )

    



import io
from typing import Union, Optional

class StorageFileBase:
    def __init__(self, mode) -> None:
        self.__readable = ("r" in mode)
        self.__writable = ("w" in mode) or ("a" in mode)

    def append(self, data : bytes):
        if self.__writable:
            raise NotImplementedError()
        else:
            raise RuntimeError("StorageFile is not writable")

    def readinto(self, __buffer) -> Optional[int]:
        if self.__readable:
            raise NotImplementedError()
        else:
            raise RuntimeError("StorageFile is not readable")
    
    def flush(self):
        raise NotImplementedError()
    
    def close(self):
        raise NotImplementedError()


class StorageBase:
    def __init__(self):
        pass
    
    def open(self, path : str, mode : str, begin : int = None, end : int = None) -> StorageFileBase:
        raise NotImplementedError()
    
    def filesize(self, path : str) -> Union[int, None]:
        raise NotImplementedError()
    
    def put(self, path : str, data : Union[bytes, io.IOBase]) -> None:
        raise NotImplementedError()
    
    def readfile(self, path : str, chunk_size = 128 * 1024) -> bytes:
        fp = self.open(path, "r")

        buffer = io.BytesIO()
        memview = bytearray(chunk_size)
        while True:
            lw = fp.readinto(memview)
            if lw == 0 or lw is None:
                break
            buffer.write( memview[:lw] )
        fp.close()
        return buffer.getvalue()



from typing import Any
from ..abc import Serializer
import orjson

class JSONSerializer(Serializer):
    def serialize(self, x : Any) -> bytes:
        return orjson.dumps(x)
    
    def deserialize(self, x : bytes) -> Any:
        return orjson.loads(x)



from ..abc import Serializer

class NoSerializer(Serializer):
    def serialize(self, x : bytes) -> bytes:
        if not isinstance(x, bytes):
            raise TypeError("Serializer is required if you want to write Object data")
        return x
    
    def deserialize(self, x : bytes) -> bytes:
        return x


from ..row import RowDataset
from ..abc import DatasetIterator

class SequentialIterator(DatasetIterator):
    def __init__(self, ds : RowDataset, epoch: int):
        self.__ds = ds
        self.__ds.seek(0)
    
    def next(self):
        try:
            return self.__ds.read()
        except EOFError:
            raise StopIteration()



import kara_storage

def handle_obj(args):
    kwargs = {}
    if args.app_key is not None:
        kwargs["app_key"] = args.app_key
    if args.app_secret is not None:
        kwargs["app_secret"] = args.app_secret
    storage = kara_storage.KaraStorage(args.url, **kwargs)
    
    if args.action == "load":
        if args.version is None:
            args.version = "latest"
        storage.load_directory(args.namespace, args.key, args.path, args.version)
    elif args.action == "save":
        storage.save_directory(args.namespace, args.key, args.path, args.version)
    


import os
from typing import Any, List
import threading
import queue
import hashlib
from ..abc import StorageBase
from tqdm import tqdm

class ObjectDataset:
    def __init__(self, storage : StorageBase, num_workers=4, chunk_size = 128 * 1024):
        self.__storage = storage
        self.num_workers = num_workers
        self.chunk_size = chunk_size
    
    def __calc_local_file_hash(self, path):

        md5 = hashlib.md5()
        sha1 = hashlib.sha1()
        
        fp = open(path, "rb")
        while True:
            data = fp.read(self.chunk_size)
            if len(data) == 0:
                break
            md5.update(data)
            sha1.update(data)
        fp.close()
        return md5.hexdigest(), sha1.hexdigest()


    def __download_thread(self, q : queue.Queue, path_prefix, remote_prefix, q_bar : queue.Queue):
        while True:
            try:
                v = q.get_nowait()
            except queue.Empty:
                break
            local_path = os.path.join(path_prefix, *v["path"])
            if os.path.exists(local_path):
                # local file exists, check hash
                md5, sha1 = self.__calc_local_file_hash(local_path)
                if v["md5"] == md5 and v["sha1"] == sha1:
                    # same file, ignore
                    q_bar.put( v["size"] )
                    continue
            
            # download remote file
            f_remote = self.__storage.open(remote_prefix + v["file"], "r")

            # makedir for local file
            os.makedirs(os.path.dirname(local_path), exist_ok=True)
            f_local = open(local_path, "wb")
            
            # trunk downloader
            memview = bytearray(self.chunk_size)
            while True:
                lw = f_remote.readinto(memview)
                if lw == 0 or lw is None:
                    break
                q_bar.put(lw)
                f_local.write( memview[:lw] )
            f_remote.close()
            f_local.close()

    def download(self, data_prefix : str, file_info, local_path : str, progress_bar=True) -> None:
        # normalize data_prefix
        if not data_prefix.endswith("/"):
            data_prefix = data_prefix + "/"
        
        q_file = queue.Queue()
        total_file_size = 0
        for it in file_info:
            q_file.put(it)
            total_file_size += it["size"]
        
        q_bar = queue.Queue()
        thds = [
            threading.Thread(target=self.__download_thread, args=(q_file, local_path, data_prefix, q_bar), daemon=True) 
                for _ in range(self.num_workers)
        ]
        thd_bar = threading.Thread(target=self.__progress_bar, args=(q_bar, progress_bar), kwargs={"total": total_file_size, "desc": "Downloading", "unit": "B", "unit_scale": True}, daemon=True)
        thd_bar.start()
        for thd in thds:
            thd.start()
        
        for thd in thds:
            thd.join()
        q_bar.put(None)
        thd_bar.join()
        
        path_set = set()
        for it in file_info:
            str_path = "/".join(it["path"])
            path_set.add(str_path)
        
        for fullname, relative_path in self.__search_in_file(local_path, []):
            str_path = "/".join(relative_path)
            if str_path not in path_set:
                os.unlink(fullname)

    def __progress_bar(self, q_upd : queue.Queue, progress_bar : bool, *args, **kwargs):
        if progress_bar:
            with tqdm(*args, **kwargs) as pbar:
                while True:
                    v = q_upd.get()
                    if v is None:
                        break
                    pbar.update(v)
        else:
            while True:
                v = q_upd.get()
                if v is None:
                    break

    
    def __search_in_file(self, local_path, path : List[str]):
        for fname in os.listdir(local_path):
            
            fullname = os.path.join(local_path, fname)
            relative_path = (path + [fname]).copy()

            if os.path.isdir(fullname):
                for it in self.__search_in_file(fullname, relative_path):
                    yield it
            else:
                yield (fullname, relative_path)
    
    def __hash_thread(self, q_in : queue.Queue, q_out : queue.Queue, q_bar : queue.Queue):
        while True:
            try:
                local_path, relative_path = q_in.get_nowait()
            except queue.Empty:
                break
            md5, sha1 = self.__calc_local_file_hash(local_path)
            q_out.put({
                "md5": md5,
                "sha1": sha1,
                "file": md5 + "_" + sha1,
                "path": relative_path,
                "size": os.stat(local_path).st_size
            })
            q_bar.put(1)
    
    def __upload_thread(self, q_in : queue.Queue, remote_path : str, q_bar : queue.Queue):
        while True:
            try:
                info = q_in.get_nowait()
            except queue.Empty:
                break
            
            if self.__storage.filesize(remote_path + info["file"]) is None:
                # file not exists
                self.__storage.put( remote_path + info["file"] , open(info["local_path"], "rb") )
            q_bar.put( info["size"] )
            
                
    def upload(self, data_prefix : str, local_path : str, progress_bar=True) -> Any:
        # normalize data_prefix
        if not data_prefix.endswith("/"):
            data_prefix = data_prefix + "/"
        
        q_in = queue.Queue()
        q_out = queue.Queue()
        q_bar = queue.Queue()

        cnt_files = 0
        for it in self.__search_in_file(local_path, []):
            q_in.put(it)
            cnt_files += 1

        # start hash thread
        thds = [
            threading.Thread(target=self.__hash_thread, args=(q_in, q_out, q_bar), daemon=True) 
                for _ in range(self.num_workers)
        ]
        thd_bar = threading.Thread(target=self.__progress_bar, args=(q_bar, progress_bar), kwargs={"total": cnt_files, "desc": "Hashing"}, daemon=True)
        thd_bar.start()
        for thd in thds:
            thd.start()

        for thd in thds:
            thd.join()
        q_bar.put(None)
        thd_bar.join()

        # end hash threads
        file_info = []
        while not q_out.empty():
            file_info.append( q_out.get() )
        
        assert q_in.empty()
        assert q_out.empty()
        assert q_bar.empty()

        # filter identical files

        total_file_size = 0
        file_content_set = set()
        for info in file_info:
            if info["file"] not in file_content_set:
                file_content_set.add( info["file"] )
                fpath = os.path.join(local_path, *info["path"])
                total_file_size += os.stat(fpath).st_size
                q_in.put({
                    "file": info["file"],
                    "local_path": fpath,
                    "size": info["size"]
                })
        
        # start upload threads
        thds = [
            threading.Thread(target=self.__upload_thread, args=(q_in, data_prefix, q_bar), daemon=True)
                for _ in range(self.num_workers)
        ]
        
        thd_bar = threading.Thread(target=self.__progress_bar, args=(q_bar, progress_bar), kwargs={"total": total_file_size, "desc": "Uploading", "unit": "B", "unit_scale": True}, daemon=True)
        thd_bar.start()
        for thd in thds:
            thd.start()
        
        for thd in thds:
            thd.join()
        q_bar.put(None)
        thd_bar.join()
        return file_info
        
        
        
        




import io
from typing import Union
from ..abc import StorageBase, StorageFileBase
import oss2

class OSSFile(StorageFileBase):
    pass

class OSSReadableFile(OSSFile):
    def __init__(self, fp : oss2.models.GetObjectResult) -> None:
        super().__init__("r")
        self.__fp = fp

    def readinto(self, __buffer):
        v = self.__fp.read(len(__buffer))
        __buffer[:len(v)] = v
        return len(v)
        
    def flush(self):
        return
    
    def close(self):
        self.__fp.close()
    
class OSSAppendableFile(StorageFileBase):
    def __init__(self, bucket : oss2.Bucket, path) -> None:
        super().__init__("a")
        self.__path = path
        self.__bucket = bucket
        try:
            self.__infile_offset = self.__bucket.get_object_meta(path).content_length
        except oss2.exceptions.NoSuchKey:
            self.__infile_offset = 0
        

    def append(self, data : Union[bytes, memoryview]):
        if isinstance(data, memoryview):
            data = data.tobytes()
        elif isinstance(data, bytes):
            pass
        else:
            raise TypeError("Invalid data type: %s (require bytes/memoryview)" % type(data))
        self.__bucket.append_object(self.__path, self.__infile_offset, data)
        self.__infile_offset += len(data)
    
    def flush(self):
        return
    
    def close(self):
        return

class OSSStorage(StorageBase):
    def __init__(self, bucket : str, endpoint : str, app_key : str, app_secret : str):
        self.__bucket_name = bucket
        self.__endpoint = endpoint
        self.auth = oss2.Auth(app_key, app_secret)
        self.bucket = oss2.Bucket(self.auth, endpoint, bucket)
        
    
    def open(self, path, mode, begin=None, end=None) -> OSSFile:
        if path.startswith("/"):
            path = path[1:]
        if mode == "r":
            if end is not None:
                return OSSReadableFile(self.bucket.get_object(path, byte_range=(begin, end - 1))) 
            else:
                return OSSReadableFile(self.bucket.get_object(path, byte_range=(begin, None)))
        elif mode == "a":
            # self.bucket is a stateful object with a session of signature
            
            return OSSAppendableFile(oss2.Bucket(self.auth, self.__endpoint, self.__bucket_name), path)
        else:
            raise ValueError("Unknown mode: `%s`" % mode)
        
    
    def filesize(self, path : str):
        if path.startswith("/"):
            path = path[1:]
        try:
            return self.bucket.get_object_meta(path).content_length
        except oss2.exceptions.NoSuchKey:
            return None
    
    def put(self, path: str, data: Union[bytes, io.IOBase] ):
        if path.startswith("/"):
            path = path[1:]
        self.bucket.put_object(path, data)



from typing import TYPE_CHECKING
from ..row import RowDataset

if TYPE_CHECKING:
    from .base import KaraPytorchDatasetBase

def make_torch_dataset(ds : RowDataset, shuffle=False, auto_distributed=True, **kwargs) -> 'KaraPytorchDatasetBase':
    
    import torch
    import torch.distributed
    from .base import KaraPytorchDatasetBase
    from .iter import SequentialIterator
    from .shuffle import ShuffleIterator

    if torch.distributed.is_initialized() and auto_distributed:
        rank = torch.distributed.get_rank()
        size = torch.distributed.get_world_size()

        total_length = ds.size()

        ds.slice_(total_length * rank // size, total_length // size)
    if shuffle:
        ret = KaraPytorchDatasetBase(ds,  ShuffleIterator, **kwargs)
    else:
        ret = KaraPytorchDatasetBase(ds, SequentialIterator, **kwargs)
    return ret



from .trunk import TrunkController
from ..abc import StorageBase, Dataset
import struct
import io

class RawDataset(Dataset):
    def __init__(self, storage : StorageBase, prefix : str, mode : str, buffer_size : int = 1024 * 1024, **kwargs) -> None:
        if not prefix.endswith("/"):
            prefix = prefix + "/"

        self.__closed = True
        self.__mode = mode
        self.__writable = ("w" in mode)
        self.__readable = ("r" in mode)
        self.__index_controller = TrunkController(storage, prefix + "index/" , mode=mode, **kwargs)
        self.__data_controller = TrunkController(storage, prefix + "data/" , mode=mode, **kwargs)
        
        if self.__readable:
            self.__index_reader = io.BufferedReader(self.__index_controller, buffer_size=buffer_size)
            self.__data_reader = io.BufferedReader(self.__data_controller, buffer_size=buffer_size)
        if self.__writable:
            self.__index_writer = io.BufferedWriter(self.__index_controller, buffer_size=buffer_size)
            self.__data_writer = io.BufferedWriter(self.__data_controller, buffer_size=buffer_size)
        
        self.__closed = False
        self.__last_read_pos = 0

        self.__real_data_size = self.__data_controller.size
        self.__tell = 0
        self.__size = self.__index_controller.size // 8
    
    def __del__(self):
        self.close()
    
    @property
    def closed(self):
        return self.__closed
    
    def close(self):
        if not self.__closed:
            if self.__readable:
                self.__index_reader.close()
                self.__data_reader.close()
            if self.__writable:
                self.__index_writer.close()
                self.__data_writer.close()
            self.__closed = True
    
    def flush(self):
        if self.__writable:
            if self.__closed:
                raise RuntimeError("Dataset closed")
            self.__index_writer.flush()
            self.__data_writer.flush()
    
    def write(self, data : bytes):
        if self.__closed:
            raise RuntimeError("Dataset closed")
        if not self.__writable:
            raise RuntimeError("Dataset not writable in mode `%s`" % self.__mode)

        self.__data_writer.write(data)
        self.__real_data_size += len(data)
        self.__size += 1
        self.__index_writer.write( struct.pack("Q", self.__real_data_size) )

    
    def read(self) -> bytes:
        if self.__closed:
            raise RuntimeError("Dataset closed")
        if not self.__readable:
            raise RuntimeError("Dataset not readable in mode `%s`" % self.__mode)
        if self.__tell == self.__size:
            return None
        v = self.__index_reader.read(8)
        if v is None:
            return None
        if len(v) != 8:
            raise RuntimeError("Dataset is broken at index offset %d, got length %d" % (self.__tell * 8, len(v)))
        cur_read_pos = struct.unpack("Q", v)[0]
        length = cur_read_pos - self.__last_read_pos
        ret = self.__data_reader.read(length)
        if len(ret) != length:
            raise RuntimeError("Dataset is broken at data offset %d ~ %d" % (self.__last_read_pos, cur_read_pos))
        self.__last_read_pos = cur_read_pos
        self.__tell += 1
        return ret
        

    
    def seek(self, offset : int, whence : int) -> int:
        if self.__closed:
            raise RuntimeError("Dataset closed")
        if not self.__readable:
            raise RuntimeError("Dataset not seekable in mode `%s`" % self.__mode)

        nw_pos = None
        if whence == io.SEEK_SET:
            nw_pos = offset
        elif whence == io.SEEK_CUR:
            nw_pos = self.__tell + offset
        elif whence == io.SEEK_END:
            nw_pos = self.__size - offset
        else:
            raise ValueError("Invalid whence: %d" % whence)
        if nw_pos < 0:
            nw_pos = 0
        if nw_pos > self.__size:
            nw_pos = self.__size
        if nw_pos > 0:
            self.__index_reader.seek((nw_pos - 1) * 8, io.SEEK_SET)
            self.__last_read_pos = struct.unpack("Q", self.__index_reader.read(8))[0]
        else:
            self.__index_reader.seek(0, io.SEEK_SET)
            self.__last_read_pos = 0

        self.__data_reader.seek(self.__last_read_pos, io.SEEK_SET)
        self.__tell = nw_pos

        return self.__tell
            
    def pread(self, offset : int) -> bytes:
        if self.__closed:
            raise RuntimeError("Dataset closed")
        if not self.__readable:
            raise RuntimeError("Dataset not readable in mode `%s`" % self.__mode)
        if offset >= self.__size:
            raise IndexError("Offset %d is out of range [0, %d)" % (offset ,self.__size))

        if offset > 0:
            bf = self.__index_controller.pread((offset - 1) * 8, 16)
            if len(bf) != 16:
                raise RuntimeError("Dataset is broken at index offset %d, go length %d" % ((offset - 1) * 8, len(bf)))
            last_pos = struct.unpack("Q", bf[:8])[0]
            curr_pos = struct.unpack("Q", bf[8:])[0]
        else:
            bf = self.__index_controller.pread(0, 8)
            if len(bf) != 8:
                raise RuntimeError("Dataset is broken at index offset %d" % 0)
            last_pos = 0
            curr_pos = struct.unpack("Q", bf)[0]
            
        ret = self.__data_controller.pread( last_pos, curr_pos - last_pos )
        if len(ret) != curr_pos - last_pos:
            raise RuntimeError("Dataset is broken at data offset %d ~ %d" % (last_pos, curr_pos))
        return ret
    
    def size(self) -> int:
        return self.__size
    
    def tell(self) -> int:
        return self.__tell
    


import io
from kara_storage.row.proxy import RowDatasetProxy
import multiprocessing
from multiprocessing.connection import Connection
from typing import Any, Dict, Generator, Optional
from .dataset import RawDataset
from ..abc import StorageBase, Dataset, Serializer
from ..serialization import JSONSerializer
import threading
from multiprocessing.reduction import ForkingPickler

class RowDataset(Dataset):
    """
    RowDataset adds multi-threading, multi-processing and slicing capabilities to RawDataset.
    """
    def __init__(self, 
            storage: StorageBase, 
            prefix: str, 
            mode: str, 
            serialization : Serializer = None,
            start : int = None,
            length : int = None,
            **kwargs
        ) -> None:

        # record initial parameters
        self.__storage = storage
        self.__prefix = prefix
        self.__mode = mode
        self.__kwargs = kwargs
        self.__writable = ("w" in mode)
        self.__readable = ("r" in mode)

        # create RawDataset
        self.__ds = RawDataset(storage, prefix, mode, **kwargs)

        # initialize serializer
        if serialization is None:
            serialization = JSONSerializer()
        self.__serialization = serialization

        # init lock
        self.__lock = threading.Lock()
        
        # init ipc
        self.__ipc_lock = threading.Lock()
        self.__ipc_server = None
        self.__ipc_client_cnt = 0
        self.__ipc_serial = 0
        self.__ipc_pipes : Dict[int, Connection] = {}
        self.__ipc_pipes_back : Dict[int, Connection] = {}

        # init slice
        if self.__readable:
            if start is None:
                start = 0
            if length is None:
                length = self.__ds.size() - start
            
            self.__begin = start
            self.__end = start + length
            if self.__end > self.__ds.size():
                self.__end = self.__ds.size()
            self.__length = self.__end - self.__begin
            self.__tell = 0
            self.__ds.seek(self.__begin, io.SEEK_SET)
        else:
            self.__length = self.__ds.size()
    
    def __read_raw(self) -> Optional[bytes]:
        if not self.__readable:
            raise RuntimeError("Dataset is not readable")
        if self.__ds.closed:
            raise RuntimeError("Dataset is closed")

        with self.__lock:
            if self.__tell == self.__length:
                return None
            self.__tell += 1
            return self.__ds.read()
    
    def __pread_raw(self, offset : int) -> Optional[bytes]:
        if not self.__readable:
            raise RuntimeError("Dataset is not readable")
        if self.__ds.closed:
            raise RuntimeError("Dataset is closed")
        if offset >= self.__length:
            return None

        with self.__lock:
            return self.__ds.pread(offset + self.__begin)
    
    def __write_raw(self, data : bytes):
        if not self.__writable:
            raise RuntimeError("Dataset is not writable")
        if self.__ds.closed:
            raise RuntimeError("Dataset is closed")

        with self.__lock:
            self.__ds.write(data)
    
    @property
    def closed(self):
        with self.__lock:
            return self.__ds.closed
    
    def close(self):
        with self.__lock:
            self.__ds.close()
    
    def flush(self):
        with self.__lock:
            self.__ds.flush()
    
    def write(self, data : Any):
        byte_data = self.__serialization.serialize(data)
        self.__write_raw(byte_data)
    
    def read(self) -> Any:
        byte_ret = self.__read_raw()
        if byte_ret is None:
            raise EOFError()
        return self.__serialization.deserialize(byte_ret)

    def seek(self, offset : int, whence : int = io.SEEK_SET) -> int:
        if not self.__readable:
            raise RuntimeError("Dataset is not readable")
        if self.__ds.closed:
            raise RuntimeError("Dataset is closed")

        if whence == io.SEEK_SET:
            ds_offset = self.__begin + offset
        elif whence == io.SEEK_CUR:
            ds_offset = self.__tell + self.__begin + offset
        elif whence == io.SEEK_END:
            ds_offset = self.__end - offset
        if ds_offset < self.__begin:
            ds_offset = self.__begin
        if ds_offset > self.__end:
            ds_offset = self.__end

        with self.__lock:
            self.__tell = ds_offset - self.__begin
            return self.__ds.seek(ds_offset, io.SEEK_SET)
            
    def pread(self, offset : int) -> Any:
        byte_ret = self.__pread_raw(offset)
        if byte_ret is None:
            raise EOFError()
        return self.__serialization.deserialize(byte_ret)
    
    def size(self) -> int:
        with self.__lock:
            return self.__length
    
    def tell(self) -> int:
        if not self.__readable:
            raise RuntimeError("Dataset is not readable")
        with self.__lock:
            return self.__tell
    
    def __ipc_main(self):
        import select
        while True:
            with self.__ipc_lock:
                if self.__ipc_client_cnt == 0:
                    # remove server if all clients exists
                    self.__ipc_server = None
                    assert len(self.__ipc_pipes) == 0
                    self.__ipc_serial = 0
                    break

                pipes = list(self.__ipc_pipes.values())
            
            readable, _, _ = select.select(pipes, [], [])
            for pipe in readable:
                cmd = pipe.recv()
                try:
                    if cmd["op"] == "closed":
                        pipe.send({
                            "code": 0,
                            "data": self.closed
                        })
                    elif cmd["op"] == "close":
                        self.close()
                        pipe.send({"code": 0})
                    elif cmd["op"] == "flush":
                        self.flush()
                        pipe.send({"code": 0})
                    elif cmd["op"] == "write":
                        self.__write_raw(cmd["data"])
                        pipe.send({"code": 0})
                    elif cmd["op"] == "read":
                        pipe.send({
                            "code": 0,
                            "data": self.__read_raw()
                        })
                    elif cmd["op"] == "seek":
                        pipe.send({
                            "code": 0,
                            "data": self.seek(*cmd["data"])
                        })
                    elif cmd["op"] == "pread":
                        pipe.send({
                            "code": 0,
                            "data": self.__pread_raw(cmd["data"])
                        })
                    elif cmd["op"] == "size":
                        pipe.send({
                            "code": 0,
                            "data": self.size()
                        })
                    elif cmd["op"] == "size":
                        pipe.send({
                            "code": 0,
                            "data": self.tell()
                        })
                    elif cmd["op"] == "exit":
                        serial_id = cmd["data"]
                        with self.__ipc_lock:
                            self.__ipc_client_cnt -= 1
                            self.__ipc_pipes[serial_id].close()
                            self.__ipc_pipes_back[serial_id].close()

                            del self.__ipc_pipes[serial_id]
                            del self.__ipc_pipes_back[serial_id]
                    else:
                        raise ValueError("Unknown cmd: %s" % cmd)
                except Exception as e:
                    pipe.send({
                        "code": 1,
                        "data": e
                    })

    
    def _reduce_dataset(self):
        with self.__ipc_lock:
            self.__ipc_client_cnt += 1
            self.__ipc_serial += 1

            serial_id = self.__ipc_serial

            p1, p2 = multiprocessing.Pipe()
            self.__ipc_pipes[serial_id] = p1
            self.__ipc_pipes_back[serial_id] = p2
            if self.__ipc_server is None:
                self.__ipc_server = threading.Thread(target=self.__ipc_main)
                self.__ipc_server.start()
        return RowDatasetProxy, (serial_id, p2, self.__serialization)
    
    def __len__(self) -> int:
        return self.__length
    
    def __iter__(self) -> Generator[Any, None, None]:
        while True:
            try:
                v = self.read()
            except EOFError:
                break
            yield v
    
    def __getitem__(self, key : int) -> Any:
        if not isinstance(key, int):
            raise TypeError("Dataset index must be int")
        try:
            return self.pread(key)
        except EOFError:
            raise IndexError("Index `%d` is out of range" % key)
    
    def slice(self, start : int = 0, length : int = None):
        if not self.__readable:
            raise RuntimeError("Dataset is not readable")
        if self.__ds.closed:
            raise RuntimeError("Dataset is closed")

        if length is None:
            length = self.__length - start
        return RowDataset(
            self.__storage, 
            self.__prefix, 
            self.__mode, 
            self.__serialization, 
            self.__begin + start,
            length,
            **self.__kwargs
        )
    
    def slice_(self, start : int = 0, length : int = None):
        if not self.__readable:
            raise RuntimeError("Dataset is not readable")
        if self.__ds.closed:
            raise RuntimeError("Dataset is closed")

        if length is None:
            length = self.__length - start
        
        with self.__lock:
            self.__begin += start
            self.__end = self.__begin + length

            assert self.__end <= self.__ds.size()

            self.__length = self.__end - self.__begin
            self.__tell = 0
            self.__ds.seek(self.__begin, io.SEEK_SET)
    
    def __del__(self):
        self.close()

def register_forking():
    ForkingPickler.register(RowDataset, RowDataset._reduce_dataset)

register_forking()
    



import kara_storage
import blessed
import signal
import json

widths = [
    (126,    1), (159,    0), (687,     1), (710,   0), (711,   1), 
    (727,    0), (733,    1), (879,     0), (1154,  1), (1161,  0), 
    (4347,   1), (4447,   2), (7467,    1), (7521,  0), (8369,  1), 
    (8426,   0), (9000,   1), (9002,    2), (11021, 1), (12350, 2), 
    (12351,  1), (12438,  2), (12442,   0), (19893, 2), (19967, 1),
    (55203,  2), (63743,  1), (64106,   2), (65039, 1), (65059, 0),
    (65131,  2), (65279,  1), (65376,   2), (65500, 1), (65510, 2),
    (120831, 1), (262141, 2), (1114109, 1),
]
 
def char_width( o ):
    global widths
    if o == 0xe or o == 0xf:
        return 0
    for num, wid in widths:
        if o <= num:
            return wid
    return 1

def handle_row(args):
    
    kwargs = {}
    if args.app_key is not None:
        kwargs["app_key"] = args.app_key
    if args.app_secret is not None:
        kwargs["app_secret"] = args.app_secret
    storage = kara_storage.KaraStorage(args.url, **kwargs)
    ds = storage.open_dataset(
        args.namespace, 
        args.key,
        mode="r", 
        version="latest" if args.version is None else args.version,
        buffer_size = 128 * 1024
    )
    ds.seek( args.begin )

    term = blessed.Terminal()

    
    idx_ = args.begin
    with term.fullscreen(), term.cbreak(), term.hidden_cursor():
        view_port_offx = 0
        view_port_offy = 0
        text = json.dumps(ds.read(), indent="    ", ensure_ascii=False).splitlines(keepends=False)

        def draw(*args):
            row = term.height
            col = term.width

            with term.location(0, 0):
                top_bar = " Index: %d" % idx_
                print( term.black_on_darkkhaki( top_bar + (" " * (col - len(top_bar))) ) )

            with term.location(0, 1):
                for i in range(row - 2):
                    if i + view_port_offx < len(text) and i + view_port_offx >= 0:
                        line = text[i + view_port_offx]
                        st_pos = 0
                        st_idx = 0
                        while st_idx < len(line) and st_pos < view_port_offy:
                            st_pos += char_width( ord(line[st_idx]) )
                            st_idx += 1
                        if st_idx >= len(line):
                            print(" " * col)
                            continue
                        st_pos -= view_port_offy
                        print_line = " " * st_pos
                        while st_idx < len(line) and st_pos + char_width( ord(line[st_idx]) ) < col:
                            print_line += line[st_idx]
                            st_pos += char_width( ord(line[st_idx]) )
                            st_idx += 1
                        print_line += " " * (col - st_pos)
                        print(print_line)
                    else:
                        print(" " * col)

            with term.location(0, row - 1):
                hint_line = "[N]: Next          [T]: To the top    [B]: To the bottom [Q]: Quit"
                print( term.black_on_darkkhaki( hint_line + (" " * (col - len(hint_line))) ) , end="")

        draw()
        signal.signal(signal.SIGWINCH, draw)
        while True:
            inp = term.inkey()
            if inp.is_sequence:
                if inp.name.lower() == "key_left":
                    view_port_offy += 2
                    draw()
                elif inp.name.lower() == "key_right":
                    view_port_offy -= 2
                    draw()
                elif inp.name.lower() == "key_up":
                    view_port_offx += 1
                    draw()
                elif inp.name.lower() == "key_down":
                    view_port_offx -= 1
                    draw()
            else:
                if inp.lower() == "q":
                    break
                elif inp.lower() == "n":
                    try:
                        text = json.dumps(ds.read(), indent="    ", ensure_ascii=False).splitlines(keepends=False)
                    except EOFError:
                        break
                    view_port_offx = 0
                    view_port_offy = 0
                    idx_ += 1
                    draw()
                elif inp.lower() == "t":
                    view_port_offx = 0
                    view_port_offy = 0
                    draw()
                elif inp.lower() == "b":
                    view_port_offx = len(text) - (term.height - 2)
                    view_port_offy = 0
                    draw()
                




import io
from typing import Any, Generator
from ..abc import Dataset, Serializer
import multiprocessing.connection
import atexit

class RowDatasetProxy(Dataset):
    def __init__(self, serial_id : int, pipe : multiprocessing.connection.Connection, serialization : Serializer) -> None:
        self.__serial_id = serial_id
        self.__pipe = pipe
        self.__serialization = serialization
        atexit.register(self.__handle_exit)

    @property
    def closed(self):
        self.__pipe.send({"op": "closed"})
        ret = self.__pipe.recv()
        if ret["code"] == 0:
            return ret["data"]
        elif ret["code"] == 1:
            raise ret["data"]
        else:
            raise ValueError("Unknown response from dataset server: %s" % ret)
    
    def close(self):
        self.__pipe.send({"op": "close"})
        ret = self.__pipe.recv()
        if ret["code"] == 0:
            return
        elif ret["code"] == 1:
            raise ret["data"]
        else:
            raise ValueError("Unknown response from dataset server: %s" % ret)
    
    def flush(self):
        self.__pipe.send({"op": "flush"})
        ret = self.__pipe.recv()
        if ret["code"] == 0:
            return
        elif ret["code"] == 1:
            raise ret["data"]
        else:
            raise ValueError("Unknown response from dataset server: %s" % ret)
    
    def write(self, data : Any):
        byte_data = self.__serialization.serialize(data)
        
        self.__pipe.send({"op": "write", "data": byte_data})
        ret = self.__pipe.recv()
        if ret["code"] == 0:
            return
        elif ret["code"] == 1:
            raise ret["data"]
        else:
            raise ValueError("Unknown response from dataset server: %s" % ret)
    
    def read(self) -> Any:
        self.__pipe.send({"op": "read"})
        ret = self.__pipe.recv()
        if ret["code"] == 0:
            if ret["data"] is None:
                raise EOFError()
            return self.__serialization.deserialize(ret["data"])
        elif ret["code"] == 1:
            raise ret["data"]
        else:
            raise ValueError("Unknown response from dataset server: %s" % ret)

    def seek(self, offset : int, whence : int = io.SEEK_SET) -> int:
        self.__pipe.send({"op": "seek", "data": (offset, whence)})
        ret = self.__pipe.recv()
        if ret["code"] == 0:
            return ret["data"]
        elif ret["code"] == 1:
            raise ret["data"]
        else:
            raise ValueError("Unknown response from dataset server: %s" % ret)
            
    def pread(self, offset : int) -> Any:
        self.__pipe.send({"op": "pread", "data": offset})
        ret = self.__pipe.recv()
        if ret["code"] == 0:
            if ret["data"] is None:
                raise EOFError()
            return self.__serialization.deserialize(ret["data"])
        elif ret["code"] == 1:
            raise ret["data"]
        else:
            raise ValueError("Unknown response from dataset server: %s" % ret)
    
    def size(self) -> int:
        self.__pipe.send({"op": "size"})
        ret = self.__pipe.recv()
        if ret["code"] == 0:
            return ret["data"]
        elif ret["code"] == 1:
            raise ret["data"]
        else:
            raise ValueError("Unknown response from dataset server: %s" % ret)
    
    def tell(self) -> int:
        self.__pipe.send({"op": "tell"})
        ret = self.__pipe.recv()
        if ret["code"] == 0:
            return ret["data"]
        elif ret["code"] == 1:
            raise ret["data"]
        else:
            raise ValueError("Unknown response from dataset server: %s" % ret)
    
    def __handle_exit(self):
        self.__pipe.send({"op": "exit", "data": self.__serial_id})
    
    def __del__(self):
        atexit.unregister(self.__handle_exit)
        self.__handle_exit()
    
    def __len__(self) -> int:
        return self.size()
    
    def __iter__(self) -> Generator[Any, None, None]:
        while True:
            try:
                v = self.read()
            except EOFError:
                break
            yield v
    
    def __getitem__(self, key : int) -> Any:
        if not isinstance(key, int):
            raise TypeError("Dataset index must be int")
        try:
            return self.pread(key)
        except EOFError:
            raise IndexError("Index `%d` is out of range" % key)


import io
import os
from typing import Union
from ..abc import StorageBase, StorageFileBase

class LocalFile(StorageFileBase):
    def __init__(self, fp : io.RawIOBase, mode) -> None:
        super().__init__(mode)

        self.__fp = fp

    def append(self, data : bytes):
        rest_length = len(data)
        while rest_length > 0:
            wrt_len = self.__fp.write(data[-rest_length:])
            rest_length -= wrt_len
        

    def readinto(self, __buffer):
        return self.__fp.readinto(__buffer)
    
    def flush(self):
        self.__fp.flush()
    
    def close(self):
        self.__fp.close()

class LocalFileStorage(StorageBase):
    def __init__(self):
        pass
    
    def open(self, path, mode, begin=None, end=None) -> LocalFile:
        if mode == "r":
            fp = open(path, "rb")
            if begin is not None:
                fp.seek(begin, io.SEEK_SET)
        elif mode == "a":
            os.makedirs(os.path.dirname( os.path.abspath(path) ), exist_ok=True)
            fp = open(path, "ab")
        else:
            raise ValueError("Unknown mode: `%s`" % mode)
        return LocalFile(fp, mode)
        
    
    def filesize(self, path):
        if not os.path.exists(path):
            return None
        return os.stat(path).st_size
    
    def put(self, path: str, data: Union[bytes, io.RawIOBase, io.BufferedIOBase]):
        # make dir for new file
        os.makedirs(os.path.dirname( os.path.abspath(path) ), exist_ok=True)

        if isinstance(data, bytes):
            open(path, "wb").write(data)
        else:
            buf = bytearray(4 * 1024 * 1024)
            fout = open(path, "wb")
            while True:
                lw = data.readinto(buf)
                if lw is None or lw == 0:
                    break
                fout.write(buf[:lw])
            fout.close()


import random

from ..abc import DatasetIterator
from ..row import RowDataset


class ShuffleIterator(DatasetIterator):
    def __init__(self, ds : RowDataset, epoch: int, seed : int = 0, buffer_size = 4096) -> None:
        self.__g = random.Random(epoch + seed)
        self.__ds = ds
        
        self.__ds.seek(0)
        
        self.__pool = []
        for i in range(buffer_size):
            try:
                self.__pool.append(self.__ds.read())
            except EOFError:
                break
    
    def next(self):
        if len(self.__pool) == 0:
            raise StopIteration()

        idx = int(self.__g.random() * len(self.__pool))
        ret = self.__pool[idx]

        try:
            nwval = self.__ds.read()
        except EOFError:
            self.__pool[idx] = self.__pool[-1]
            self.__pool.pop()
        else:
            self.__pool[idx] = nwval
        return ret


from typing import Any
from .storage import StorageBase
import io

class Dataset:    
    @property
    def closed(self):
        raise NotImplementedError()
    
    def close(self):
        raise NotImplementedError()
    
    def flush(self):
        raise NotImplementedError()
    
    def write(self, data : Any):
        raise NotImplementedError()
    
    def read(self) -> Any:
        raise NotImplementedError()

    def seek(self, offset : int, whence : int = io.SEEK_SET) -> int:
        raise NotImplementedError()
            
    def pread(self, offset : int) -> Any:
        raise NotImplementedError()
    
    def size(self) -> int:
        raise NotImplementedError()
    
    def tell(self) -> int:
        raise NotImplementedError()
    


import argparse

def bind_obj(parser : argparse.ArgumentParser):
    parser.add_argument("url", help="KARA Storage location", type=str)
    parser.add_argument("action", help="actions: load or save directory", type=str, choices=["load", "save"])
    parser.add_argument("namespace", help="namespace", type=str)
    parser.add_argument("key", help="key", type=str)
    parser.add_argument("-v", "--version", type=str, default=None, help="version")
    parser.add_argument("path", help="local path", type=str)

    parser.add_argument("--app-key", type=str, default=None, help="OSS app key")
    parser.add_argument("--app-secret", type=str, default=None, help="OSS app secret")

    from ..funcs import handle_obj
    parser.set_defaults(func=handle_obj)



import os
import setuptools
from tools import get_requirements, get_readme, get_version

path = os.path.dirname(os.path.abspath(__file__))

version = get_version()

open( os.path.join(path, "kara_storage", "version.py"), "w" ).write('version = "%s"' % version)



def main():
    setuptools.setup(
        name="kara_storage",
        version=version,
        author="a710128",
        author_email="qbjooo@qq.com",
        description="Kara Storage SDK",
        long_description=get_readme(),
        long_description_content_type="text/markdown",
        url="https://git.thunlp.vip/kara/kara-row-storage",
        packages=setuptools.find_packages(exclude=("tools",)),
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Programming Language :: C++"
        ],
        python_requires=">=3.6",
        setup_requires=["wheel"],
        scripts=["scripts/kara_storage"],
        install_requires=get_requirements()
    )

if __name__ == "__main__":
    main()



import io
import os
from typing import Union
from ..abc import StorageBase, StorageFileBase
import urllib
import urllib.request
import urllib.error

class HTTPFile(StorageFileBase):
    def __init__(self, fp) -> None:
        super().__init__("r")
        self.__fp = fp

    def append(self, data : bytes):
        raise RuntimeError("HTTP/HTTPS files are read-only")
        

    def readinto(self, __buffer):
        return self.__fp.readinto(__buffer)
    
    def flush(self):
        return
    
    def close(self):
        self.__fp.close()

class HTTPStorage(StorageBase):
    def __init__(self, url_prefix : str, headers = {}):
        if not url_prefix.endswith("/"):
            url_prefix = url_prefix + "/"
        self.__url_prefix = url_prefix
        self.__custom_headers = headers
    
    def __range_to_str(self, v):
        if v is None:
            return ""
        return str(v)
    
    def open(self, path : str, mode, begin=None, end=None) -> HTTPFile:
        if path.startswith("/"):
            path = path[1:]
        if mode == "r":
            if begin is None and end is None:
                req = urllib.request.Request(self.__url_prefix + path, headers=self.__custom_headers)
            else:
                if end is not None:
                    end = end - 1
                req = urllib.request.Request(self.__url_prefix + path, headers={
                    **self.__custom_headers,
                    "Range": "bytes=%s-%s" % (
                        self.__range_to_str(begin),
                        self.__range_to_str(end)
                    )
                })
            resp = urllib.request.urlopen(req)
            if begin is not None or end is not None:
                if 'content-range' not in resp.headers:
                    raise RuntimeError("Storage server seems doesn't support range header")
            return HTTPFile(resp)
        elif mode == "a":
            raise ValueError("HTTP/HTTPS Storages are read-only")
        else:
            raise ValueError("Unknown mode: `%s`" % mode)
        
    
    def filesize(self, path : str):
        if path.startswith("/"):
            path = path[1:]
        try:
            resp = urllib.request.urlopen(urllib.request.Request(self.__url_prefix + path, headers=self.__custom_headers))
        except urllib.error.HTTPError as e:
            if e.code != 404:
                raise RuntimeError("Unexpected response code: %d" % e.code)
            return None
        if not "content-length" in resp.headers:
            raise RuntimeError("Storage server does not support `Content-Length` header")

        return int(resp.headers["content-length"])
    
    def put(self, path: str, data: Union[bytes, io.RawIOBase, io.BufferedIOBase]):
        raise ValueError("HTTP/HTTPS Storages are read-only")


from typing import Dict, Generator, Any, Type
import torch.utils.data as data
from ..row import RowDataset
from ..abc import DatasetIterator
from multiprocessing.connection import Connection
import multiprocessing as mp
import threading
import atexit
import logging

logger = logging.getLogger(__name__)

class KaraPytorchDatasetProxy(data.IterableDataset):
    def __init__(self, serial_id, pipe):
        self.__serial_id = serial_id
        self.__pipe = pipe
        atexit.register(self.__handle_exit)
        logger.debug("KARA Pytorch proxy %d created" % self.__serial_id)
    
    def __iter__(self):
        while True:
            try:
                self.__pipe.send({
                    "op": 0,
                })
                data = self.__pipe.recv()
                if data["code"] == 0:
                    pass
                elif data["code"] == 1:
                    raise data["data"]
                else:
                    raise ValueError("Unknown IPC server response %s" % data)

            except StopIteration:
                break
            else:
                yield data["data"]
    
    def __handle_exit(self):
        self.__pipe.send({
            "op": 1,
            "data": self.__serial_id
        })
        logger.debug("KARA Pytorch proxy %d exited" % self.__serial_id)
    
    def __del__(self):
        atexit.unregister(self.__handle_exit)
        self.__handle_exit()
    

class KaraPytorchDatasetBase(data.IterableDataset):
    def __init__(self, dataset : RowDataset, iter_tool : Type[DatasetIterator], **kwargs) -> None:
        self.__ds = dataset
        self.__iter_tool = iter_tool
        self.__kwargs = kwargs
        self.__iter = None

        # read lock
        self.__lock = threading.Lock()

        # ipc server
        self.__ipc_server = None
        self.__ipc_client_cnt = 0
        self.__ipc_serial = 0
        self.__ipc_pipes : Dict[int, Connection] = {}
        self.__ipc_pipes_back : Dict[int, Connection] = {}
        self.__ipc_lock = threading.Lock()

        self.set_epoch(0)

    
    def set_epoch(self, epoch_num : int):
        with self.__lock:
            if self.__iter is not None:
                del self.__iter
                self.__iter = None
            self.__iter = self.__iter_tool(self.__ds, epoch_num, **self.__kwargs)
    
    def __read_next(self):
        if self.__iter is None:
            raise StopIteration()
        with self.__lock:
            return self.__iter.next()

    def __iter__(self) -> Generator[Any, None, None]:
        while True:
            try:
                v = self.__read_next()
            except StopIteration:
                break
            else:
                yield v
    
    def __len__(self) -> int:
        raise len(self.__ds)
    
    def __ipc_main(self):
        import select
        logger.debug("KARA Pytorch IPC server started")
        while True:
            with self.__ipc_lock:
                if self.__ipc_client_cnt == 0:
                    # remove server if all clients exists
                    assert len(self.__ipc_pipes) == 0
                    assert len(self.__ipc_pipes_back) == 0

                    self.__ipc_server = None
                    self.__ipc_serial = 0
                    break

                pipes = list(self.__ipc_pipes.values())
            
            readable, _, _ = select.select(pipes, [], [])
            for pipe in readable:
                cmd = pipe.recv()
                try:
                    if cmd["op"] == 0:
                        # read
                        pipe.send({
                            "code": 0,
                            "data": self.__read_next()
                        })
                    elif cmd["op"] == 1:
                        # exit
                        serial_id = cmd["data"]
                        logger.debug("KARA Pytorch IPC server recived proxy %d exited" % serial_id)
                        with self.__ipc_lock:
                            self.__ipc_client_cnt -= 1
                            self.__ipc_pipes[serial_id].close()
                            self.__ipc_pipes_back[serial_id].close()
                            del self.__ipc_pipes[serial_id]
                            del self.__ipc_pipes_back[serial_id]
                    else:
                        logger.debug("KARA Pytorch received invalid ipc cmd: %s", cmd)
                        raise ValueError("Unknown IPC proto")
                except Exception as e:
                    pipe.send({
                        "code": 1,
                        "data": e,
                    })
        # ipc server exits
        logger.debug("KARA Pytorch IPC server stoped")
                    
    
    def __reduce__(self):
        with self.__ipc_lock:
            self.__ipc_client_cnt += 1
            self.__ipc_serial += 1

            serial_id = self.__ipc_serial

            p1, p2 = mp.Pipe()
            self.__ipc_pipes[serial_id] = p1
            self.__ipc_pipes_back[serial_id] = p2
            if self.__ipc_server is None:
                self.__ipc_server = threading.Thread(target=self.__ipc_main)
                self.__ipc_server.start()
        return KaraPytorchDatasetProxy, (serial_id, p2)
        


import io
from kara_storage.abc.serializer import Serializer
from typing import Any
from urllib.parse import urlparse
import os
import warnings
import json
from ..row import RowDataset
from ..object import ObjectDataset

class KaraStorage:
    def __init__(self, url, **kwargs) -> None:
        uri = urlparse(url)
        if uri.scheme == "file":
            path = ""
            if uri.netloc == "":
                path = uri.path
            else:
                path = os.path.abspath(uri.netloc) + uri.path
            from ..backend.file import LocalFileStorage
            self.__prefix = path
            self.__storage = LocalFileStorage()
        elif uri.scheme == "oss":
            path =  uri.path.split("/")
            from ..backend.oss import OSSStorage
            if "use_ssl" in kwargs and kwargs["use_ssl"]:
                self.__storage = OSSStorage(path[1], "https://" + uri.netloc, kwargs["app_key"], kwargs["app_secret"])
            else:
                self.__storage = OSSStorage(path[1], "http://" + uri.netloc, kwargs["app_key"], kwargs["app_secret"])
            self.__prefix = "/".join(path[2:])    
        elif uri.scheme == "http" or uri.scheme == "https":
            from ..backend.http import HTTPStorage
            headers = {}
            if "headers" in kwargs:
                headers = kwargs["headers"]
            self.__storage = HTTPStorage( uri.scheme + "://" + uri.netloc, headers=headers)
            self.__prefix = uri.path
        else:
            raise ValueError("Unknown scheme `%s`" % uri.scheme)
        
        self.__object_dataset = ObjectDataset(self.__storage)
        
        if not self.__prefix.endswith("/"):
            self.__prefix = self.__prefix + "/"
        
    
    
    def __get_meta(self, storage_type : str, namespace : str, key : str):
        version_path = self.__prefix + "%s/%s/%s/meta.json" % (storage_type, namespace, key)
        if self.__storage.filesize(version_path) is None:
            raise FileNotFoundError("Dataset not exists")

        return json.loads(self.__storage.readfile(version_path).decode("utf-8"))
    
    def __get_row_meta(self, namespace : str, key : str):
        return self.__get_meta("row", namespace, key)
    
    def __get_object_meta(self, namespace : str, key : str):
        return self.__get_meta("obj", namespace, key)
    
    def __put_meta(self, storage_type : str, namespace : str, key : str, meta : Any):
        version_path = self.__prefix + "%s/%s/%s/meta.json" % (storage_type, namespace, key)
        self.__storage.put( version_path, json.dumps(meta).encode("utf-8") )
    
    def __put_row_meta(self, namespace : str, key : str, meta : Any):
        self.__put_meta("row", namespace, key, meta)
    
    def __put_object_meta(self, namespace : str, key : str, meta : Any):
        self.__put_meta("obj", namespace, key, meta)
    
    def open_dataset(self, 
        namespace : str, key : str, mode : str = "r", version="latest", 
        serialization : Serializer = None, **kwargs
    ) -> RowDataset:

        version = str(version)
        try:
            config = self.__get_row_meta(namespace, key)
        except FileNotFoundError:
            if "w" not in mode:
                raise ValueError("Dataset not exists")
            config = {
                "latest": None,
                "versions": [],
                "api": 2,
            }
        
        if version == "latest":
            if config["latest"] is None:
                raise ValueError("No available version found in dataset `%s`. Please specify a version if you want to create a new dataset" % key)
            version = config["latest"]

        if "w" in mode:
            config["latest"] = version
            if version not in config["versions"]:
                config["versions"].append(version)
            self.__put_row_meta(namespace, key, config)
            
        if "r" in mode:
            if version not in config["versions"]:
                raise ValueError("Dataset version `%s` not found in dataset `%s`" % (version, key))
        
        return RowDataset(self.__storage, self.__prefix + "row/%s/%s/%s/" % (namespace, key, version), mode, serialization=serialization, **kwargs)

    def open(self, *args, **kwargs) -> RowDataset:
        warnings.warn(
            "KaraStorage.open is deprecated, and will be removed in the future.\n" +
            "Please use KaraStorage.open_dataset instead."
        )
        return self.open_dataset(*args, **kwargs)

    def load_directory(self, namespace : str, key : str, local_path : str, version = "latest", progress_bar=True):
        
        try:
            config = self.__get_object_meta(namespace, key)
        except FileNotFoundError:
            raise ValueError("Dataset not exists")
        
        version = str(version)
        if version == "latest":
            if config["latest"] is None:
                raise ValueError("No available version found in object storage `%s`." % key)
            version = config["latest"]

        if version not in config["versions"]:
            raise ValueError("Object version `%s` not found in storage `%s`" % (version, key))
        
        version_info = json.loads(self.__storage.readfile(
            self.__prefix + "obj/%s/%s/vers/%s.json" % (namespace, key, version)
        ).decode("utf-8"))
        self.__object_dataset.download(
            self.__prefix + "obj/%s/%s/data/" % (namespace, key), 
            version_info,
            local_path,
            progress_bar=progress_bar
        )
        return version
    
    def save_directory(self, namespace : str, key : str, local_path : str, version = None, progress_bar=True) -> str:
        
        try:
            config = self.__get_object_meta(namespace, key)
        except FileNotFoundError:
            config = {
                "latest": None,
                "versions": [],
                "api": 2
            }
        if version is None:
            # auto generate version
            cnt = 0
            while ("%d" % cnt) in config["versions"]:
                cnt += 1
            version = '%d' % cnt
        
        version = str(version)
        if version == "latest":
            if config["latest"] is None:
                raise ValueError("No available version found in object storage `%s`." % key)
            version = config["latest"]
        
        config["latest"] = version
        if version not in config["versions"]:
            config["versions"].append(version)
        self.__put_object_meta(namespace, key, config)

        version_info = self.__object_dataset.upload(
            self.__prefix + "obj/%s/%s/data/" % (namespace, key),
            local_path,
            progress_bar=progress_bar
        )
        self.__storage.put(
            self.__prefix + "obj/%s/%s/vers/%s.json" % (namespace, key, version),
            json.dumps( version_info ).encode("utf-8")
        )
        return version

    

    


        


